package industries.doublehelix.pwste;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.stream.Stream;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFormattedTextField;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.Timer;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.text.NumberFormatter;

import org.jcodec.api.awt.AWTSequenceEncoder;

public class PowerWashSimTimelapseExporter extends JFrame implements ActionListener, ChangeListener {
	private JLabel lblImage = new JLabel();//new StretchIcon("C:/Users/Akhlys/Pictures/1vLLI3A.png"));
	JSlider sliNumberOfFrames = new JSlider(JSlider.HORIZONTAL, 10, 100, 100);
	JComboBox<TimeLapse> cmbTimelapses = null;
	JFormattedTextField txtDelay = null;
	JFormattedTextField txtDelayEnd = null;
	
	private JButton btnExportAll = new JButton("Export all Images");
	private JButton btnExportGif = new JButton("Export as animated Gif");
	private JButton btnExportVideo = new JButton("Export as video");
	
	private List<TimeLapse> timeLapses = new ArrayList<>();
	
	private ArrayList<byte[]> currentImagesSubset = new ArrayList<>();
	private int currentFrame = 0;
	private Timer timer = null;
	
	private File lastChooserDirectory = null;
	
	public PowerWashSimTimelapseExporter() {
		loadTimelapses();
		if (this.timeLapses.isEmpty()) return; // end it
		
		Container pane = this.getContentPane();
		((JPanel) pane).setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		JPanel mainPanel = new JPanel();
		mainPanel.setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 4;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.anchor = GridBagConstraints.WEST;
		c.insets = new Insets(0,  0,  10,  10);
		mainPanel.add(new JLabel("Available Timelapses:"), c);
		cmbTimelapses = new JComboBox<>(timeLapses.toArray(new TimeLapse[0]));
		cmbTimelapses.setSelectedIndex(-1);
		cmbTimelapses.addActionListener(this);
		//restrictMaximumHeight(cmbTimelapses);
		c.gridy = 1;
		c.weightx = 1;
		mainPanel.add(cmbTimelapses, c);
		

		c.gridy = 2;
		c.weightx = 0;
		mainPanel.add(new JLabel("Preview:"), c);
		
		c.gridy = 3;
		c.weighty = 1;
		c.weightx = 1;
		c.fill = GridBagConstraints.BOTH;
		//lblImage.setMaximumSize(new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE));
		//lblImage.setPreferredSize(new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE));
		mainPanel.add(lblImage, c);

		c.weighty = 0;
		c.weightx = 0;
		c.gridy = 4;
		c.fill = GridBagConstraints.HORIZONTAL;
		mainPanel.add(new JLabel("Number of frames to export (for gif/video):"), c);
		c.gridy = 5;
		c.weightx = 1;
		sliNumberOfFrames.setMajorTickSpacing(10);
		sliNumberOfFrames.setMinorTickSpacing(1);
		sliNumberOfFrames.setPaintTicks(true);
		sliNumberOfFrames.setPaintLabels(true);
		sliNumberOfFrames.addChangeListener(this);
		mainPanel.add(sliNumberOfFrames, c);
		
		JPanel delayPanel = new JPanel();
		delayPanel.setLayout(new BoxLayout(delayPanel, BoxLayout.X_AXIS));
		delayPanel.add(new JLabel("Delay between frames (in milliseconds):"));
		delayPanel.add(Box.createHorizontalStrut(10));
		NumberFormat intFormat = NumberFormat.getIntegerInstance();
		intFormat.setGroupingUsed(false);
		NumberFormatter numberFormatter = new NumberFormatter(intFormat);
		numberFormatter.setValueClass(Integer.class);
		//numberFormatter.setAllowsInvalid(false);
		numberFormatter.setMinimum(10);
		numberFormatter.setMaximum(9999);
		txtDelay = new JFormattedTextField(numberFormatter);
		txtDelay.setColumns(4);
		txtDelay.setValue(100);
		txtDelay.setMaximumSize(txtDelay.getPreferredSize());
		txtDelay.addActionListener(this);
		delayPanel.add(txtDelay);
		delayPanel.add(Box.createHorizontalStrut(10));
		delayPanel.add(new JLabel("(hit enter to update preview)"));
		//delayPanel.add(Box.createHorizontalGlue());
		//restrictMaximumHeight(delayPanel);
		c.gridy = 6;
		c.weightx = 0;
		mainPanel.add(delayPanel, c);
		
		JPanel delayEndPanel = new JPanel();
		delayEndPanel.setLayout(new BoxLayout(delayEndPanel, BoxLayout.X_AXIS));
		delayEndPanel.add(new JLabel("Pause at the end (in milliseconds):"));
		delayEndPanel.add(Box.createHorizontalStrut(10));

		numberFormatter = new NumberFormatter(intFormat);
		numberFormatter.setValueClass(Integer.class);
		numberFormatter.setMinimum(10);
		numberFormatter.setMaximum(99999);
		txtDelayEnd = new JFormattedTextField(numberFormatter);
		txtDelayEnd.setColumns(5);
		txtDelayEnd.setValue(1000);
		txtDelayEnd.setMaximumSize(txtDelay.getPreferredSize());
		txtDelayEnd.addActionListener(this);
		delayEndPanel.add(txtDelayEnd);
		delayEndPanel.add(Box.createHorizontalStrut(10));
		delayEndPanel.add(new JLabel("(hit enter to update preview)"));
		//delayPanel.add(Box.createHorizontalGlue());
		//restrictMaximumHeight(delayPanel);
		c.gridy = 7;
		mainPanel.add(delayEndPanel, c);

		btnExportAll.addActionListener(this);
		btnExportGif.addActionListener(this);
		btnExportVideo.addActionListener(this);
		
		c.gridy = 8;
		c.gridwidth = 1;
		mainPanel.add(btnExportAll, c);
		c.gridx = 1;
		mainPanel.add(btnExportGif, c);
		c.gridx = 2;
		mainPanel.add(btnExportVideo, c);
		
		this.add(mainPanel);
		
		sliNumberOfFrames.setEnabled(false);
		txtDelay.setEnabled(false);
		txtDelayEnd.setEnabled(false);
		btnExportAll.setEnabled(false);
		btnExportGif.setEnabled(false);
		btnExportVideo.setEnabled(false);
		
		this.setSize(800, 800);
		this.setLocationRelativeTo(null);
		this.setVisible(true);
		this.setTitle("Akhlys' Powerwash Simulator Timelapse Exporter");
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	}
	
	private void loadTimelapses() {
		Path mainDir = Paths.get(System.getenv("USERPROFILE"));
		mainDir = mainDir.resolve("AppData\\LocalLow\\FuturLab\\PowerWash Simulator\\SaveData\\0");
		
		if (!Files.exists(mainDir)) {
			JOptionPane.showMessageDialog(this, "Could not find save directory. please select your save dir. It should have been in " + mainDir.toString(), "Error", JOptionPane.ERROR_MESSAGE);
			JFileChooser chooser = new JFileChooser();
			chooser.setDialogTitle("Select your save dir");
		    chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		    chooser.setAcceptAllFileFilterUsed(false);
		    //chooser.setCurrentDirectory(new java.io.File("."));
			boolean directoryFound = false;
			while (!directoryFound) {
				if (chooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
					Path saveDirSelected = Paths.get(chooser.getSelectedFile().getAbsolutePath());
					if (!Files.exists(saveDirSelected.resolve("Player.sav"))) {
						JOptionPane.showMessageDialog(this, "That dir contains no player.sav and doesn't seem to be your save dir. Try again (or hit cancel to exit)", "Error", JOptionPane.ERROR_MESSAGE);
					} else {
						directoryFound = true;
						mainDir = saveDirSelected;
					}
				} else {
					System.exit(0);
				}
			}
		}
		
		List<Path> dirs = new ArrayList<>(); // I hate working within streams because of exceptions etc
		try (Stream<Path> filePathStream = Files.list(mainDir)) {
			filePathStream.forEach(filePath -> {
				if (Files.isDirectory(filePath) && filePath.getFileName().toString().toLowerCase().endsWith("timelapse")) {
					dirs.add(filePath);
				}
			});
		} catch (IOException e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(this, "Could not read Save directory, sorry. Cause: " + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}
		
		// for sorting
		Map<String, TimeLapse> timelapsesMap = new TreeMap<>();
		for (Path dir : dirs) {
			try {
				TimeLapse lapse = new TimeLapse(dir);
				lapse.load();
				timelapsesMap.put(lapse.getProcessedName(), lapse);
			} catch (Exception e) {
				System.out.println("Skipping dir " + dir.toString() + " because:");
				e.printStackTrace();
			}
		}
		this.timeLapses.addAll(timelapsesMap.values());
	}
	
	public static void main(String[] args) {
		new PowerWashSimTimelapseExporter();
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == cmbTimelapses) {
			TimeLapse timeLapse = (TimeLapse)cmbTimelapses.getSelectedItem();
			int selectedFrames = sliNumberOfFrames.getValue();
			int numFrames = timeLapse.getImages().size();
			sliNumberOfFrames.setMaximum(numFrames);
			if (selectedFrames > numFrames) sliNumberOfFrames.setValue(numFrames);
			updateAndRestartPreview();
			
			sliNumberOfFrames.setEnabled(true);
			txtDelay.setEnabled(true);
			txtDelayEnd.setEnabled(true);
			btnExportAll.setEnabled(true);
			btnExportGif.setEnabled(true);
			btnExportVideo.setEnabled(true);
		} else if (e.getSource() == txtDelay) {
			if (this.timer != null) this.timer.setDelay((Integer)txtDelay.getValue());
		} else if (e.getSource() == timer) {
			if (currentImagesSubset.size() - 1 < currentFrame) {
				currentFrame = 0;
			}
			if (currentImagesSubset.size() -2 == currentFrame) {
				timer.setDelay((Integer)txtDelayEnd.getValue());
			}
			if (currentImagesSubset.size() -1 == currentFrame) {
				timer.setDelay((Integer)txtDelay.getValue());
			}
			lblImage.setIcon(new StretchIcon(currentImagesSubset.get(currentFrame)));
			++currentFrame;
		} else if (e.getSource() == btnExportAll) {
			JFileChooser chooser = new JFileChooser(); 
		    //chooser.setCurrentDirectory(new java.io.File("."));
		    if (lastChooserDirectory != null) chooser.setCurrentDirectory(lastChooserDirectory);
		    chooser.setDialogTitle("Select a directory to save the files in");
		    chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		    chooser.setAcceptAllFileFilterUsed(false);
		    if (chooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
		    	this.lastChooserDirectory = chooser.getSelectedFile();
		        Path destiny = chooser.getSelectedFile().toPath();
		        TimeLapse timeLapse = (TimeLapse)cmbTimelapses.getSelectedItem();
		        String filename = timeLapse.getRawName().toLowerCase();
		        List<byte[]> imageList = timeLapse.getImages();
		        try {
			        for (int i = 0; i < imageList.size(); ++i) {
			        	byte[] image = imageList.get(i);
			        	Files.write(destiny.resolve(filename + i + ".jpg"), image);
			        }
		        } catch (Exception ex) {
		        	JOptionPane.showMessageDialog(this, "Could not save everything. Cause: " + ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
		        	ex.printStackTrace();
		        }
		    }
		} else if (e.getSource() == btnExportGif) {
			JFileChooser chooser = new JFileChooser(); 
		    chooser.setDialogTitle("Save gif");
		    if (lastChooserDirectory != null) chooser.setCurrentDirectory(lastChooserDirectory);
		    chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		    chooser.setAcceptAllFileFilterUsed(false);
		    chooser.setFileFilter(new FileNameExtensionFilter("Gif Files", "gif"));
		    
		    if (chooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
		    	try {
		    		this.getContentPane().setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			        String strPath = chooser.getSelectedFile().getAbsolutePath();
			        if (!strPath.toLowerCase().endsWith(".gif")) strPath += ".gif";
			        this.lastChooserDirectory = new File(strPath).getParentFile();
			    	//Path destiny = Paths.get(strPath);
			        
			        
			    	AnimatedGifEncoder encoder = new AnimatedGifEncoder();
			    	encoder.setRepeat(0);
			    	encoder.start(strPath);
			    	encoder.setDelay((Integer)txtDelay.getValue());
			    	for (int i = 0; i < currentImagesSubset.size() - 1; ++i) {
				    	encoder.addFrame(getBufferedImgFromByteArray(currentImagesSubset.get(i)));
			    	}
			    	encoder.setDelay((Integer)txtDelayEnd.getValue());
			    	encoder.addFrame(getBufferedImgFromByteArray(currentImagesSubset.get(currentImagesSubset.size() - 1)));
			    	encoder.finish();
		    	} catch (Exception ex) {
		    		JOptionPane.showMessageDialog(this, "Could not save gif. Cause: " + ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
		    		ex.printStackTrace();
		    	}
	    		this.getContentPane().setCursor(Cursor.getDefaultCursor());
		    	
		    	if ((Integer)txtDelay.getValue() < 100) {
		    		JOptionPane.showMessageDialog(this, "Gif saved. Note you selected a very short delay and gifs aren't good at handling that so it may run slower than you intended.", "Information", JOptionPane.INFORMATION_MESSAGE);
		    	}
		    }
		} else if (e.getSource() == btnExportVideo) {
			JFileChooser chooser = new JFileChooser(); 
		    chooser.setDialogTitle("Save mp4");
		    if (lastChooserDirectory != null) chooser.setCurrentDirectory(lastChooserDirectory);
		    chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		    chooser.setAcceptAllFileFilterUsed(false);
		    chooser.setFileFilter(new FileNameExtensionFilter("MP4 Files", "mp4"));
		    if (chooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
	    		this.getContentPane().setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		        String strPath = chooser.getSelectedFile().getAbsolutePath();
		        if (!strPath.toLowerCase().endsWith(".mp4")) strPath += ".mp4";
		        this.lastChooserDirectory = new File(strPath).getParentFile();
				try {
					int fps = 1000/(Integer)txtDelay.getValue();
					AWTSequenceEncoder enc = AWTSequenceEncoder.createSequenceEncoder(new File(strPath), fps);
					for (int i = 0; i < currentImagesSubset.size() - 1; ++i) {
						enc.encodeImage(getBufferedImgFromByteArray(currentImagesSubset.get(i)));
					}
					// add the last frame to fill up the last frame delay
					BufferedImage endImage = getBufferedImgFromByteArray(currentImagesSubset.get(currentImagesSubset.size() - 1));
					if ((Integer)txtDelayEnd.getValue() > (Integer)txtDelay.getValue() * 2) {
						int numRepetitions = (Integer)txtDelayEnd.getValue() / (Integer)txtDelay.getValue();
						for (int i = 0; i < numRepetitions; ++i) {
							enc.encodeImage(endImage);
						}
						System.out.println(numRepetitions);
					} else {
						enc.encodeImage(endImage);
					}
					enc.finish();
				} catch (Exception ex) {
					JOptionPane.showMessageDialog(this, "Could not save video. Cause: " + ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
		    		ex.printStackTrace();
				}
				this.getContentPane().setCursor(Cursor.getDefaultCursor());
		    }
		}
	}
	
	private BufferedImage getBufferedImgFromByteArray(byte[] imagedata) throws IOException {
		InputStream is = new ByteArrayInputStream(imagedata);
        BufferedImage newBi = ImageIO.read(is);
        is.close();
        return newBi;
	}

	@Override
	public void stateChanged(ChangeEvent e) {
		if (e.getSource() == sliNumberOfFrames) {
			if (sliNumberOfFrames.getValueIsAdjusting()) return;
			updateAndRestartPreview();
		}
	}
	
	private void updateAndRestartPreview() {
		if (timer != null) timer.stop();
		TimeLapse timeLapse = (TimeLapse)cmbTimelapses.getSelectedItem();
		int delay = (Integer)txtDelay.getValue();
		int frames = sliNumberOfFrames.getValue();
		
		currentImagesSubset.clear();
		currentFrame = 0;
		int numOfImages = timeLapse.getImages().size();
		
		for(int i = 0; i < frames; ++i) {
	        if (i == frames - 1) {
	        	currentImagesSubset.add(timeLapse.getImages().get(timeLapse.getImages().size() -1));
	        } else {
	        	currentImagesSubset.add(timeLapse.getImages().get((int)Math.floor(i * numOfImages / frames)));
	        }
	    }
		
		
//		int stepping = 0;
//		if (numOfImages <= frames) {
//			stepping = 1;
//		} else {
//			stepping = numOfImages / frames;
//		}
//		System.out.println("steppng: " + stepping);
//		for (int i = stepping - 1; i < numOfImages; i += stepping) {
//			currentImagesSubset.add(timeLapse.getImages().get(i));
//		}
		
		if (timer == null) {
			timer = new Timer(delay, this);
		} else {
			timer.setDelay(delay);
		}
		timer.setInitialDelay(delay);
		timer.start();
	}
	
}
